/* Copyright (c) 2003 Andrew Stikheev. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * $Id: statistics.c,v 1.18 2003/06/01 18:34:03 sand Exp $
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <arpa/inet.h>
#include <unistd.h>

#include "statistics.h"

int                     PutStatOut (statisticItem * , int);
static int              UpdateItem (statisticItem *);
static inline long      Hash (in_addr_t);
static inline double    Update1Weight  ( double , long );
static inline double    Update2Weight  ( double , long );

extern int		enableBlock;
extern clusterItem	*hashTable;
extern int		fdLog;
static  in_addr_t 	srcIP, dstIP;
static  u_short 	srcPort, dstPort;
////////////////////////////////////////////////////////////////////////////////
int  AddPacketToStat(char *packet)
{
	static clusterItem	*cluster;
	static signed char	i;
	static signed char	j; 		// most useless Item

	static struct ip               *ip;
	static struct tcphdr           *tcp;
	static struct udphdr           *udp;

        ip = ( struct ip * ) packet;

	if ( ip->ip_p == IPPROTO_TCP ) {
   
                tcp = ( struct tcphdr *)(packet + sizeof ( struct ip ));

		if ( tcp->th_flags & ~TH_SYN ) return 0;
		
		srcPort = tcp->th_sport;
		dstPort = tcp->th_dport;

        } else
        if ( ip->ip_p == IPPROTO_UDP ) {
		
		udp = ( struct udphdr *)(packet + sizeof( struct ip ));

		srcPort = udp->uh_sport;
		dstPort = udp->uh_dport;

	} else return 0; 

	srcIP = ip->ip_src.s_addr;
	dstIP = ip->ip_dst.s_addr;

	j  = -1;
	cluster = hashTable + Hash( srcIP );

	// try to find first item with same ip or removed by viewstat item

	for( i = 0 ; i < cluster->totalItems; i++ ) { 
		if (  cluster->item[i].srcIP == srcIP ) {
			
			j = i;
			break;

		} else if ( !cluster->item[i].srcIP && j == -1) 
			j = i;
	}

	// if cluster is not full then we add new item


	if ( j == -1 && cluster->totalItems < CLUSTER_SIZE )   
		j = cluster->totalItems++;

// cluster is full and need to add new item, so we find item with minimal weight

	if ( j == -1 ) {

		j = 0;

		for( i = 1 ; i < CLUSTER_SIZE ; i++ ) {

			if (cluster->item[j].weight > cluster->item[i].weight) {

				j = i;
			}
		}
	}

	return UpdateItem ( &cluster->item[j] );   // macros

}
///////////////////////////////////////////////////////////////////////////////
int    UpdateItem( statisticItem *cur )
{
	static struct timeval   tm;
	static long		delta;
	static int block;

	block = 0 ;  // not block packet

	gettimeofday(&tm,NULL);

	if ( cur->srcIP != srcIP ) {

		cur->srcIP              = srcIP;
		cur->lastDstIP          = dstIP;
		cur->lastDstPort        = dstPort;
		cur->lastSrcPort        = srcPort;
		cur->weight             = MIN_WEIGHT; 
		cur->lastTime           = tm.tv_sec;
	}
	else {

		delta  = tm.tv_sec - cur->lastTime;

		if ( 	cur->lastDstIP   != dstIP ||
			cur->lastDstPort != dstPort 	) {

			// macros!
			cur->weight += UPDATE_WEIGHT ( cur->weight , delta ); 

			cur->lastDstIP   = dstIP;
			cur->lastDstPort = dstPort;
		}

		if ( cur->weight >= MAX_WEIGHT ) {
		
			cur->weight = MAX_WEIGHT;
			block = 1; // this packet need to block
		} else 
		if ( cur->weight < MIN_WEIGHT )  {
			
			cur->weight = MIN_WEIGHT;
		}

		cur->lastSrcPort        = srcPort;
		cur->lastTime           = tm.tv_sec;
	}

	if ( fdLog >= 0 ) PutStatOut(cur, enableBlock ? block : 0 );

	return block;
}
///////////////////////////////////////////////////////////////////////////////
int PutStatOut(statisticItem *cur, int block )
{
	static char 		buf[256];  	// i hope it's enough
	static char 		size, stm[20];
	static time_t 		tm;
	static struct in_addr	ip_src,ip_dst;
	static char		ipSrc[16],ipDst[16];


	ip_src.s_addr = cur->srcIP;
	ip_dst.s_addr = cur->lastDstIP;

        tm = time(NULL);

	strftime(stm, sizeof(stm)-1, "%b %d %H:%M:%S", localtime(&tm));
	strcpy(ipSrc,(char *)inet_ntoa(ip_src));
	strcpy(ipDst,(char *)inet_ntoa(ip_dst));

	size=sprintf(buf,"%s :  %s:%d > %s:%d %.2f %s\n",stm, 
			ipSrc, htons(cur->lastSrcPort), 
			ipDst, htons(cur->lastDstPort),
			cur->weight,
			block ? "  blocked":  ""    );

        if (write(fdLog, buf, size) != size) {

                if (errno == EAGAIN || errno == ENOSPC || errno == EDQUOT) {

                        return 0;
                }
//                Quit("write log failed");
         }
	return 1;
}
///////////////////////////////////////////////////////////////////////////////
inline double Update1Weight ( double w , long delta )
{
        return  RATIO_A  * w / (1 + RATIO_B * delta );
}
///////////////////////////////////////////////////////////////////////////////
inline double Update2Weight ( double w , long delta )
{
	return  RATIO_A * w *  
		( (1 - delta / RATIO_B ) < -1.0 ? -1.0 : 1 - delta / RATIO_B ); 
} 
///////////////////////////////////////////////////////////////////////////////
inline long Hash(in_addr_t ip)
{
	return ip % NUM_CLUSTERS;
}
///////////////////////////////////////////////////////////////////////////////
